@@ -1,6 +1,3 @@
-
-/* (C)oderite 1997 by Jedi/Sector One <fdenis@industrie.capgemini.fr> */
-
 #ifndef INCLUDED_JNETD
 #define INCLUDED_JNETD
 
@@ -21,7 +18,15 @@ enum {
    JNETD_DEFAULTMAXPID = 65536
 };
 
+/* Default configuration file */
+
 #define JNETD_DEFAULTCFFILE "jnetd.cf"
 
+/* Domain appened to the IP address for RBL checks */
+
+/* Use the line below on a production box */
+#define JNETD_RBLMAP "rbl.maps.vix.com"
+/* Or the line below for local tests */
+/* #define JNETD_RBLMAP "in-addr.arpa" */
 
 #endif
