@@ -1,7 +1,7 @@
 #ifndef INCLUDED_NEWDIRENT
 #define INCLUDED_NEWDIRENT
 
-#if HAVE_DIRENT_
+#if HAVE_DIRENT_H
 # include <dirent.h>
 # define NAMLEN(dirent) strlen((dirent)->d_name)
 #else
