@@ -1,6 +1,3 @@
-
-/* (C)oderite Jedi/Sector One */
-
 #ifndef INCLUDED_NEWERRNO
 #define INCLUDED_NEWERRNO
 
@@ -57,4 +54,4 @@ extern void printerr(const char * const more,
 
 #endif
 
-extern char *newstrerror(const int err);
+extern const char *newstrerror(const int err);
