@@ -9,5 +9,8 @@
 #elif HAVE_REGEX_H
 # include <regex.h>
 #endif
+#ifdef HAVE_GLOB_H
+# include <glob.h>
+#endif
 
 #endif
