@@ -29,4 +29,8 @@
 # endif
 #endif
 
+#ifndef HAVE_STRCASECMP
+# define strcasecmp(A, B) stricmp(A, B)
+#endif
+
 #endif
