@@ -26,12 +26,13 @@ RETSIGTYPE sigterm(int dummy)
    int fd = 0;
 
    (void) dummy;
-   log_log(LOG_WARNING, NULL, _("Shutting down the server"));   
+   log_log(LOG_WARNING, NULL, _("Shutting down the server"));
    while (fd <= SOCKMAX_FD) {
-      close(fd);
+      socket_realclose(fd);
       fd++;
    }
-   
+   sleep((unsigned int) SOCKET_LINGER);
+   log_log(LOG_WARNING, NULL, _("Server off"));   
    exit(EXIT_SUCCESS);
 #if RETSIGTYPE != void
    return 0;
@@ -63,6 +64,7 @@ RETSIGTYPE sigchild(int dummy)
       servicecontext = (ServiceContext *) child->socket->geek;
       if (servicecontext == NULL) {
 	 log_log(LOG_ERR, NULL, _("No service context ?!"));
+	 signal_catch(SIGCHLD, sigchild);	 
 #if RETSIGTYPE != void
 	 return -1;
 #else
@@ -94,7 +96,7 @@ RETSIGTYPE sigchild(int dummy)
 	 }
       } else {
 	 log_log(LOG_WARNING, NULL,
-		 _("Process [%d] from service [%s] exited abnormally, please check [%s]"),
+		 _("Process [%d] from service [%s] exited abnormally (limits reached ?), please check [%s]"),
 		 (int) foundpid,
 		 tools_servicenametxt(servicecontext),
 		 servicecontext->run[0]);
@@ -102,7 +104,8 @@ RETSIGTYPE sigchild(int dummy)
    } else {
       log_log(LOG_DEBUG, NULL, _("Suspect SIGCHLD received"));
    }
-   
+
+   signal_catch(SIGCHLD, sigchild);
 #if RETSIGTYPE != void
    return 0;
 #endif
@@ -148,18 +151,20 @@ static int jnetd_builtin(SockEvent *event)
    return 0;
 }
 
+/* Check if a host is allowed to connect on a service, based upon regexes
+ * JNETD_OLD_ACCESS_CHECK, if defined, matches both the host name and its
+ * IP address against the same regex. This is yet deprecated. */
+
 static int checkallowed(const char **names, const char * const ip,
 			const ServiceContext *servicecontext)
 {
-   static const char *ips[] = { NULL, NULL };
+   static const char *ips[2] = { NULL, NULL };
    
    if (servicecontext == NULL || servicecontext->rx_allow == NULL) {
       return -1;
    }
-   if (names == NULL) {
-      return 0;
-   }
    ips[0] = ip;
+#ifdef JNETD_OLD_ACCESS_CHECK            
    if ((tools_namesmatchrx(names, servicecontext->rx_allow) > 0 ||
 	tools_namesmatchrx(ips, servicecontext->rx_allow) > 0) &&
        (tools_namesmatchrx(names, servicecontext->rx_allow_except) <= 0 ||
@@ -167,6 +172,15 @@ static int checkallowed(const char **names, const char * const ip,
       
       return 1;
    }
+#else
+   if ((tools_namesmatchrx(ips, servicecontext->rx_allow_ip) > 0 ||
+	tools_namesmatchrx(names, servicecontext->rx_allow) > 0) &&
+       (tools_namesmatchrx(ips, servicecontext->rx_allow_ip_except) <= 0 ||
+	tools_namesmatchrx(names, servicecontext->rx_allow_except) <= 0)) {
+      
+      return 1;
+   }
+#endif
    
    return 0;
 }
@@ -179,10 +193,8 @@ static int checkdenied(const char **names, const char * const ip,
    if (servicecontext == NULL || servicecontext->rx_deny == NULL) {
       return -1;
    }
-   if (names == NULL) {
-      return 0;
-   }
    ips[0] = ip;
+#ifdef JNETD_OLD_ACCESS_CHECK            
    if ((tools_namesmatchrx(names, servicecontext->rx_deny) > 0 ||
 	tools_namesmatchrx(ips, servicecontext->rx_allow) > 0) &&
        (tools_namesmatchrx(names, servicecontext->rx_deny_except) <= 0 ||
@@ -190,6 +202,51 @@ static int checkdenied(const char **names, const char * const ip,
       
       return 1;
    }   
+#else
+   if ((tools_namesmatchrx(ips, servicecontext->rx_deny_ip) > 0 ||
+	tools_namesmatchrx(names, servicecontext->rx_allow) > 0) &&
+       (tools_namesmatchrx(ips, servicecontext->rx_deny_ip_except) <= 0 ||
+	tools_namesmatchrx(names, servicecontext->rx_deny_except) <= 0)) {
+      
+      return 1;
+   }      
+#endif
+   
+   return 0;
+}
+
+int jnetd_setlimits(ServiceContext *servicecontext)
+{
+   if (servicecontext == NULL) {
+      return -1;
+   }
+   if (tools_setlimit(TOOLS_LIMIT_CORE, servicecontext->corelimit) < 0) {
+      return -1;
+   }
+   if (servicecontext->session_cpulimit != 0 && 
+       tools_setlimit(TOOLS_LIMIT_CPU, servicecontext->session_cpulimit) < 0) {
+      return -1;
+   }
+   if (servicecontext->session_filelimit != 0 &&
+       tools_setlimit(TOOLS_LIMIT_FSIZE, servicecontext->session_filelimit) < 0) {
+      return -1;
+   }
+   if (servicecontext->session_datalimit != 0 &&
+       tools_setlimit(TOOLS_LIMIT_DATA, servicecontext->session_datalimit) < 0) {
+      return -1;
+   }
+   if (servicecontext->session_stacklimit != 0 &&
+       tools_setlimit(TOOLS_LIMIT_STACK, servicecontext->session_stacklimit) < 0) {
+      return -1;
+   }
+   if (servicecontext->session_fdlimit != 0 &&
+       tools_setlimit(TOOLS_LIMIT_NOFILE, servicecontext->session_fdlimit) < 0) {
+      return -1;
+   }
+
+   if (servicecontext->session_timelimit > 0) {
+      alarm((unsigned int) servicecontext->session_timelimit);
+   }
    
    return 0;
 }
@@ -203,14 +260,18 @@ int main(int argc, const char **argv)
    ParserResult parserresult;
 
    if (argc < 2) {
-      puts(_("Usage : jnetd <configuration file>"));
+      puts(_("Usage : " JNETD_NAME " <configuration file>"));
       exit(EXIT_FAILURE);
    }
    tools_cleanenv();
+#ifdef DEBUG
+   puts(_("Debug mode enabled : not launching as a daemon."));
+#else
    if (tools_initdaemon() != 0) {
-      puts(_("Can't start daemon"));
+      puts(_("Can't spawn the daemon"));
       exit(EXIT_FAILURE);
    }   
+#endif
 #ifdef LC_ALL
    setlocale(LC_ALL, "");
 #endif
@@ -228,7 +289,8 @@ int main(int argc, const char **argv)
       
       return EXIT_FAILURE;
    }
-   
+
+   signal_catch(SIGALRM, SIG_IGN);
    signal_catch(SIGPIPE, SIG_IGN);
    signal_catch(SIGTERM, sigterm);
    signal_catch(SIGQUIT, sigterm);   
@@ -248,9 +310,8 @@ int main(int argc, const char **argv)
    puts(jnetd_version());
    printf("Socketlib release %d.%d\n",
 	  SOCKETLIB_MAJORVERSION, SOCKETLIB_MINORVERSION);
-   puts("(C)oderite 1997 by Jedi/Sector One "
-	"<fdenis@industrie.capgemini.fr>\n");
-      
+   puts("(C)oderite 1997-1998 by Jedi/Sector One <j@4u.net>\n");
+         
    if (socket_initsockets(&sockets) != 0) {
       return EXIT_FAILURE;
    }
@@ -264,18 +325,19 @@ int main(int argc, const char **argv)
 				  servicecontextpnt->endpoint,
 				  servicecontextpnt->proto,
 				  servicecontextpnt->servicename.name,
-				 servicecontextpnt->servicename.port))
+				  servicecontextpnt->servicename.port))
 	     == NULL) {
-	    log_log(LOG_ERR, NULL, _("[%s] (%s/%s) not bind <%s>"),
+	    log_log(LOG_ERR, NULL, _("[%s] (%s/%s) not bound <%s>"),
 		    tools_servicenametxt(servicecontextpnt),
 		    servicecontextpnt->endpoint, servicecontextpnt->proto,
 		    newstrerror(errno));
 	    goto nextservice;
-	 } else {
+	 } else {	    
 	    log_log(LOG_INFO, NULL, _("Activating service [%s] (%s/%s)"),
 		    tools_servicenametxt(servicecontextpnt),
 		    servicecontextpnt->endpoint, servicecontextpnt->proto);
 	 }
+	 socket_optionnodelay(servicecontextpnt->servicenetworkstatus.socket, 0);
 	 servicecontextpnt->servicenetworkstatus.socket->geek =
 	   (void *) servicecontextpnt;
 	 nextservice:
@@ -310,6 +372,36 @@ int main(int argc, const char **argv)
 	       log_log(LOG_WARNING, NULL, ipoptmsg);
 	       goto nextevent;
 	    }	    
+	    if (servicecontext->spamfilter != 0) {
+	       const int t = socket_ispammer(event->socket, event->fd, JNETD_RBLMAP);
+	       
+	       if (t > 0) {
+		  log_log(LOG_WARNING, NULL,
+			  _("[%s] is listed as a spammer : connexion to [%s] refused"),
+			  socket_getpeernameoraddr(event->socket, event->fd),
+			  tools_servicenametxt(servicecontext));
+		  socket_print(event->fd, _("\n553 Access denied - Stop spamming, lamer\r\n"));
+		  socket_hangup(event->socket, event->fd);
+		  goto nextevent;
+	       } else if (t < 0) {
+		  if (servicecontext->spamfilter == 2) {
+		     log_log(LOG_WARNING, NULL,
+			     _("An RBL check caused a DNS error. [%s] was connected to [%s] anyway, but check [%s]"),
+			     socket_getpeernameoraddr(event->socket, event->fd),
+			     tools_servicenametxt(servicecontext),
+			     JNETD_RBLMAP);
+		  } else {
+		     log_log(LOG_WARNING, NULL,
+			     _("[%s] wanted to connect to [%s] but an RBL check caused a DNS error, check [%s]"),
+			     socket_getpeernameoraddr(event->socket, event->fd),
+			     tools_servicenametxt(servicecontext),
+			     JNETD_RBLMAP);
+		     socket_print(event->fd, _("\nUnable to prove your innocence yet, please try again later\n\n"));
+		     socket_hangup(event->socket, event->fd);
+		     goto nextevent;	
+		  }
+	       }
+	    }
 	    names = socket_getpeername(event->socket, event->fd);
 	    ip = socket_getpeeraddr(event->socket, event->fd);
 	    allowed = checkallowed(names, ip, servicecontext);
@@ -366,6 +458,7 @@ int main(int argc, const char **argv)
 		       servicecontext->gateway_to_host,
 		       servicecontext->gateway_from,
 		       socket_getpeernameoraddr(event->socket, event->fd));
+	       socket_optionnodelay(clientsocket, 0);
 	       clientsocket->geek = (void *) servicecontext;
 	       clientsocket->cookie = 1;
 	       clientsocket->cookie2 = event->fd;
@@ -421,79 +514,93 @@ int main(int argc, const char **argv)
 #ifdef F_SETFD
 	 fcntl(event->fd, F_SETFD, 0);
 #endif
-	 switch (pid = fork()) {
-	  case -1 :
-	    LOG_ERRMARK(NULL, _("Fork() failed"));
-	    log_log(LOG_ERR, NULL, _("[%s] not spawned"), 
-		    tools_servicenametxt((ServiceContext *) event->socket->geek));
-	    break;
-	  case 0 :
-	    close(0);
-	    close(1);
-	    fcntl(event->fd, F_SETFD, 0);
-	    if (socket_isstream(event->socket)) {
- 	       socket_serverclose(event->socket); 
-	    }
-	    if (fd_move(0, event->fd) < 0) {
-	       log_log(LOG_DEBUG, NULL, _("File descriptors moving failure <%s>"),
-		       newstrerror(errno));
-	    }
-	    if (fd_copy(1, 0) < 0) {
-	       log_log(LOG_DEBUG, NULL, _("File descriptors copy failure <%s>"),
-		       newstrerror(errno));
-	    }
-	    close(event->fd);
-	    {
-	       ServiceContext *servicecontext = 
-		 (ServiceContext *) event->socket->geek;
-	       
+	   {
+	      int fo_eventfd = event->fd;
+	      SocketInfo fo_eventsocketinfo = event->socketinfo;
+	      ServiceContext fo_servicecontext = *((ServiceContext *) event->socket->geek);
+	      
+	      switch (pid = fork()) {
+	       case -1 :
+		 LOG_ERRMARK(NULL, _("Fork() failed"));
+		 log_log(LOG_ERR, NULL, _("[%s] not spawned"), 
+			 tools_servicenametxt((ServiceContext *) event->socket->geek));
+		 break;
+	       case 0 :
+		 close(0);
+		 close(1);
+		 fcntl(fo_eventfd, F_SETFD, 0);
+		 if (socket_isstream(event->socket)) {
+		    socket_serverclose(event->socket);
+		    tools_setucspienv(&fo_eventsocketinfo); 
+		 }
+		 if (fd_move(0, fo_eventfd) < 0) {
+		    log_log(LOG_DEBUG, NULL, _("File descriptors moving failure <%s>"),
+			    newstrerror(errno));
+		 }
+		 if (fd_copy(1, 0) < 0) {
+		    log_log(LOG_DEBUG, NULL, _("File descriptors copy failure <%s>"),
+			    newstrerror(errno));
+		 }
+		 close(fo_eventfd);
+		   {
+		      ServiceContext *servicecontext = &fo_servicecontext;
+		      
 #ifdef HAVE_CHROOT
-	       if (servicecontext->chroot != NULL) {
-		  if (chroot(servicecontext->chroot) != 0 ||
-		      chdir(PARSER_DIRECTORYSEP_STR) != 0) {
-		     log_log(LOG_ERR, NULL,
-			     _("Chroot() to [%s] failed <%s>"),
-			     servicecontext->chroot, newstrerror(errno));
-		     goto seize;
-		  }
-	       }
+		      if (servicecontext->chroot != NULL) {
+			 if (chroot(servicecontext->chroot) != 0 ||
+			     chdir(PARSER_DIRECTORYSEP_STR) != 0) {
+			    log_log(LOG_ERR, NULL,
+				    _("Chroot() to [%s] failed <%s>"),
+				    servicecontext->chroot, newstrerror(errno));
+			    goto seize;
+			 }
+		      }
 #endif
-	       if (tools_changeid(servicecontext->user,
-				  servicecontext->group) != 0) {
-		  log_log(LOG_ERR, NULL,
-			  _("Can't change user ID for service [%s]"),
-			  tools_servicenametxt(servicecontext));
-		  goto seize;
-	       }
-	       signal_catch(SIGCHLD, SIG_DFL);
-	       signal_unblock(SIGCHLD); 	       	       
-	       signal_catch(SIGTERM, SIG_DFL);
-	       signal_catch(SIGPIPE, SIG_DFL);
-	       execvp(servicecontext->run[0], servicecontext->run);
-	    }
-	    seize:
-	    if (socket_isstream(event->socket) == 0) {
-	       socket_flush(0);
-	    }
-	    puts(_("Unable to execute"));
-	    log_log(LOG_ERR, NULL, _("Unable to launch [%s]"),
-		    ((ServiceContext *) event->socket->geek)->run[0]);
-	    exit(EXIT_FAILURE);
-	  default :
-	    if (mem_resize(m_childtab, (pid + 1) * sizeof (ChildTab)) != 0) {
-	       LOG_ERRMARK(NULL, _("Children table expansion"));
-	       break;
-	    }
-	    ((ChildTab *) m_childtab->space)[pid].fd = event->fd;
-	    ((ChildTab *) m_childtab->space)[pid].socket = event->socket;
-	    ((ServiceContext *) event->socket->geek)
-	      ->servicenetworkstatus.dgramfd = event->fd;
-	    log_log(LOG_INFO, NULL, _("[%s] launched as PID [%u] for [%s]"),
-		    tools_servicenametxt((ServiceContext *) event->socket->geek),
-		    (unsigned int) pid,
-		    socket_getpeernameoraddr(event->socket, event->fd));
-	    ((ServiceContext *) event->socket->geek)->service_cur_nblimit++;
-	 }
+		      /* set limits bevor abandoning root rights */
+		      if (jnetd_setlimits(servicecontext) != 0) {
+			 log_log(LOG_ERR, NULL,
+				 _("Can't set limits for [%s] : server not launched."),
+				 tools_servicenametxt(servicecontext));
+			 goto seize;
+		      }
+		      if (tools_changeid(servicecontext->user,
+					 servicecontext->group) != 0) {
+			 log_log(LOG_ERR, NULL,
+				 _("Can't change user ID for service [%s]"),
+				 tools_servicenametxt(servicecontext));
+			 goto seize;
+		      }
+		      signal_catch(SIGCHLD, SIG_DFL);
+		      signal_unblock(SIGCHLD); 	       	       
+		      signal_catch(SIGTERM, SIG_DFL);
+		      signal_catch(SIGPIPE, SIG_DFL);
+		      signal_catch(SIGALRM, SIG_DFL);		      
+		      execv(servicecontext->run[0], servicecontext->run);
+		   }
+		 seize:
+		 if (socket_isstream(event->socket) == 0) {
+		    socket_flush(0);
+		 }
+		 puts(_("Unable to execute"));
+		 log_log(LOG_ERR, NULL, _("Unable to launch [%s]"),
+			 ((ServiceContext *) event->socket->geek)->run[0]);
+		 exit(EXIT_FAILURE);
+	       default :
+		 if (mem_resize(m_childtab, (pid + 1) * sizeof (ChildTab)) != 0) {
+		    LOG_ERRMARK(NULL, _("Children table expansion"));
+		    break;
+		 }
+		 ((ChildTab *) m_childtab->space)[pid].fd = event->fd;
+		 ((ChildTab *) m_childtab->space)[pid].socket = event->socket;
+		 ((ServiceContext *) event->socket->geek)
+		   ->servicenetworkstatus.dgramfd = event->fd;
+		 log_log(LOG_INFO, NULL, _("[%s] launched as PID [%u] for [%s]"),
+			 tools_servicenametxt((ServiceContext *) event->socket->geek),
+			 (unsigned int) pid,
+			 socket_getpeernameoraddr(event->socket, event->fd));
+		 ((ServiceContext *) event->socket->geek)->service_cur_nblimit++;
+	      }
+	   }
  	 socket_hangup(event->socket, event->fd); 
 	 break;
        case SOCKEVENT_HANGUP :
@@ -525,7 +632,7 @@ int main(int argc, const char **argv)
 	 }
 #endif
 	 if (event->socket->cookie != 0) {
-	    write(event->socket->cookie2, event->line, event->linelen);
+	    write((int) event->socket->cookie2, event->line, event->linelen);
 	 } else {
 	    jnetd_builtin(event);
 	 }
