@@ -55,7 +55,7 @@ void printerr(const char * const more,
 #endif
 }
 
-char *newstrerror(const int err)
+const char *newstrerror(const int err)
 {
    if (err == 0) {
       return "";
