@@ -27,6 +27,26 @@ static Lotus *findtoken(const char * const token)
    return NULL;
 }
 
+static long parser_atol(const char * const s)
+{
+   char *p;
+   long r;
+   errno = 0;
+   r = strtol(s, &p, 0);
+   if (errno != 0) {
+     return 0;
+   }
+   switch(*p) {
+    case 'k':
+    case 'K': r *= 1024; break;
+    case 'm':
+    case 'M': r *= 1024 * 1024; break;
+    case 'g': 
+    case 'G': r *= 1024 * 1024 * 1024; break;
+   }
+   return r;
+}
+
 static int parser_checknumber(const char * const txt)
 {
    return isdigit(*((const unsigned char *) txt));
@@ -68,21 +88,20 @@ static int parser_defaultcontext(ServiceContext * const context)
       NULL,                              /* description */
       NULL,                              /* hostname */
       { NULL, -1 },                      /* servicename */
-      0,                                 /* corelimit */
-      (const char *) PARSER_DEFAULTPROTOCOL,            /* proto */
-      (const char *) PARSER_DEFAULTENDPOINT,            /* endpoint */
+      0L,                                /* corelimit */
+      PARSER_DEFAULTPROTOCOL,            /* proto */
+      PARSER_DEFAULTENDPOINT,            /* endpoint */
       255,                               /* wait */
-      0,                                 /* session_timelimit */
+      0L,                                /* session_timelimit */
       0,                                 /* session_cpuplimit */
-      0,                                 /* session_cpulimit */
-      0,                                 /* session_filelimit */
-      0,                                 /* session_datalimit */
-      0,                                 /* session_vmemlimit */
-      0,                                 /* session_stacklimit */
+      0L,                                /* session_cpulimit */
+      0L,                                /* session_filelimit */
+      0L,                                /* session_datalimit */
+      0L,                                /* session_stacklimit */
       0,                                 /* session_fdlimit */
       0,                                 /* service_cpuplimit */
-      0,                                 /* service_cpulimit */
-      0,                                 /* service_nblimit */
+      0L,                                /* service_cpulimit */
+      0U,                                /* service_nblimit */
       0,                                 /* service_cur_nblimit */
       0,                                 /* cnx_timeout */
       60,                                /* abuse_detect_time */
@@ -97,6 +116,10 @@ static int parser_defaultcontext(ServiceContext * const context)
       NULL,                              /* rx_allow_except */      
       NULL,                              /* rx_deny */
       NULL,                              /* rx_deny_except */
+      NULL,                              /* rx_allow_ip */
+      NULL,                              /* rx_allow_ip_except */      
+      NULL,                              /* rx_deny_ip */
+      NULL,                              /* rx_deny_ip_except */      
       NULL,                              /* user */
       NULL,                              /* group */
       NULL,                              /* chroot */
@@ -110,6 +133,7 @@ static int parser_defaultcontext(ServiceContext * const context)
       0,                                 /* spawn_nblimit */
       0,                                 /* spawn_cpuplimit */
       0,                                 /* spawn_cpulimit */
+      0,                                 /* spamfilter */
       BUILTIN_NONE,                      /* builtin */
       0,                                 /* hasaction */
       {                                  /* servicenetworkstatus */
@@ -156,8 +180,8 @@ ParserResult parser_parseconfigfile(const char *file)
    ServiceContext *firstservicecontext = NULL;
    ServiceContext *lastservicecontext = NULL;
    ServiceContext tmpservicecontext[PARSER_MAXDEPTH];
-   ParserResult parserresult;
-   
+   ParserResult parserresult = { NULL, 0 };
+      
    {
       register ServiceContext *tmpservicecontextpnt = tmpservicecontext;
       size_t t = PARSER_MAXDEPTH;      
@@ -237,7 +261,10 @@ ParserResult parser_parseconfigfile(const char *file)
       }
    }
    fclose(fp);
-   parserresult.firstservicecontext = firstservicecontext;   
+   parserresult.firstservicecontext = firstservicecontext;
+   if (firstservicecontext == NULL) {
+      parserresult.error = 1;
+   }
    
    return parserresult;
 }
@@ -250,11 +277,125 @@ PRO(CORELIMIT)
       if (parser_checknumber(item) == 0) {
 	 cr0:
 	 log_log(LOG_ERR, NULL, 
-		 _("ERROR : Need a number to specify the core limit"));
+		 _("ERROR : Need a number to specify the core limit."));
 	      
 	 return -1;
       }
-      parsercontext->currentservicecontext->corelimit = (size_t) atol(item);
+      parsercontext->currentservicecontext->corelimit = (long) parser_atol(item);
+   } else {
+      goto cr0;
+   }   
+   return 0;
+}
+
+PRO(SESSION_TIMELIMIT)
+{   
+   char *item;
+   
+   if (GETNEXTOKEN) {
+      if (parser_checknumber(item) == 0) {
+	 cr0:
+	 log_log(LOG_ERR, NULL, 
+		 _("ERROR : Need a number of secondes to specify a max time limit."));
+	      
+	 return -1;
+      }
+      parsercontext->currentservicecontext->session_timelimit = (long) parser_atol(item);
+   } else {
+      goto cr0;
+   }   
+   return 0;   
+}
+
+PRO(SESSION_CPULIMIT)
+{
+   char *item;
+   
+   if (GETNEXTOKEN) {
+      if (parser_checknumber(item) == 0) {
+	 cr0:
+	 log_log(LOG_ERR, NULL, 
+		 _("ERROR : Need a number of secondes to specify a CPU usage limit."));
+	      
+	 return -1;
+      }
+      parsercontext->currentservicecontext->session_cpulimit = (long) parser_atol(item);
+   } else {
+      goto cr0;
+   }   
+   return 0;
+}
+
+PRO(SESSION_FILELIMIT)
+{
+   char *item;
+   
+   if (GETNEXTOKEN) {
+      if (parser_checknumber(item) == 0) {
+	 cr0:
+	 log_log(LOG_ERR, NULL, 
+		 _("ERROR : Need a number of bytes to specify a maximum file size."));
+	      
+	 return -1;
+      }
+      parsercontext->currentservicecontext->session_filelimit = (long) parser_atol(item);
+   } else {
+      goto cr0;
+   }   
+   return 0;
+}
+
+PRO(SESSION_DATALIMIT)
+{
+   char *item;
+   
+   if (GETNEXTOKEN) {
+      if (parser_checknumber(item) == 0) {
+	 cr0:
+	 log_log(LOG_ERR, NULL, 
+		 _("ERROR : Need a number of bytes to specify the maximum data size."));
+	      
+	 return -1;
+      }
+      parsercontext->currentservicecontext->session_datalimit = (long) parser_atol(item);
+   } else {
+      goto cr0;
+   }   
+   return 0;
+}
+
+PRO(SESSION_STACKLIMIT)
+{
+   char *item;
+   
+   if (GETNEXTOKEN) {
+      if (parser_checknumber(item) == 0) {
+	 cr0:
+	 log_log(LOG_ERR, NULL, 
+		 _("ERROR : Need a number of bytes to specify the maximum stack size."));
+	      
+	 return -1;
+      }
+      parsercontext->currentservicecontext->session_stacklimit = (long) parser_atol(item);
+   } else {
+      goto cr0;
+   }   
+   return 0;
+}
+
+PRO(SESSION_FDLIMIT)
+{
+   char *item;
+   
+   if (GETNEXTOKEN) {
+      if (parser_checknumber(item) == 0) {
+	 cr0:
+	 log_log(LOG_ERR, NULL, 
+		 _("ERROR : Need a number : the maximum wanted number of descriptors."));
+	      
+	 return -1;
+      }
+      parsercontext->currentservicecontext->session_fdlimit = (int) atoi(item);
    } else {
       goto cr0;
    }   
@@ -391,18 +532,18 @@ PRO(PROTO)
    {
       typedef struct DefaultAssociates_ {
 	 const char *proto;
-	 char *endpoint;
+	 const char *endpoint;
 	 const unsigned char wait;
       } DefaultAssociates;
       static const DefaultAssociates defaultassociates[] = {
-	 { (const char *) "ip", (const char *) "raw", 1 },
-	 { (const char *) "icmp", (const char *) "raw", 1 },
-	 { (const char *) "ggp", (const char *) "stream", 0 },
-	 { (const char *) "tcp", (const char *) "stream", 0 },
-	 { (const char *) "egp", (const char *) "stream", 0 },
-	 { (const char *) "pup", (const char *) "seqpacket", 0 },
-	 { (const char *) "udp", (const char *) "dgram", 1 },
-	 { (const char *) "rdp", (const char *) "dgram", 1 }
+	 { "ip", "raw", 1 },
+	 { "icmp", "raw", 1 },
+	 { "ggp", "stream", 0 },
+	 { "tcp", "stream", 0 },
+	 { "egp", "stream", 0 },
+	 { "pup", "seqpacket", 0 },
+	 { "udp", "dgram", 1 },
+	 { "rdp", "dgram", 1 }
       };
       register const DefaultAssociates *defaultassociatespnt = defaultassociates;
       size_t t = (sizeof defaultassociates) / (sizeof defaultassociates[0]);
@@ -624,6 +765,134 @@ PRO(DENY_EXCEPT)
    return 0;
 }
 
+PRO(ALLOW_IP)
+{
+   char *item;
+   ServiceContext *currentservicecontext =
+     parsercontext->currentservicecontext;
+   regex_t *comp;
+   int err;
+   
+   if (GETNEXTOKEN) {
+      if ((comp = (regex_t *) malloc(sizeof *comp)) == NULL) {
+	 log_log(LOG_ERR, NULL,
+		 _("No more memory to store the allowed IP addresses"));
+	 
+	 return -1;
+      }
+      if ((err = regcomp(comp, item, REG_EXTENDED | REG_ICASE)) != 0) {
+	 char errbuf[LINE_MAX + 1];
+	 
+	 regerror(err, comp, errbuf, sizeof LINE_MAX);
+	 log_log(LOG_ERR, NULL,
+		 _("Suspicious regex : [%s] <%s>"), item, errbuf);
+	 
+	 return -1;
+      }
+      currentservicecontext->rx_allow_ip = comp;
+   } else {
+      currentservicecontext->rx_allow_ip = NULL;
+   }
+
+   return 0;
+}
+
+PRO(ALLOW_IP_EXCEPT)
+{
+   char *item;
+   ServiceContext *currentservicecontext =
+     parsercontext->currentservicecontext;
+   regex_t *comp;
+   int err;
+   
+   if (GETNEXTOKEN) {
+      if ((comp = (regex_t *) malloc(sizeof *comp)) == NULL) {
+	 log_log(LOG_ERR, NULL,
+		 _("No more memory to store the excepted allowed IP addresses"));
+	 
+	 return -1;
+      }
+      if ((err = regcomp(comp, item, REG_EXTENDED | REG_ICASE)) != 0) {
+	 char errbuf[LINE_MAX + 1];
+	 
+	 regerror(err, comp, errbuf, sizeof LINE_MAX);
+	 log_log(LOG_ERR, NULL,
+		 _("Suspicious regex : [%s] <%s>"), item, errbuf);
+	 
+	 return -1;
+      }
+      currentservicecontext->rx_allow_ip_except = comp;
+   } else {
+      currentservicecontext->rx_allow_ip_except = NULL;
+   }
+
+   return 0;
+}
+
+PRO(DENY_IP)
+{
+   char *item;
+   ServiceContext *currentservicecontext =
+     parsercontext->currentservicecontext;
+   regex_t *comp;
+   int err;
+   
+   if (GETNEXTOKEN) {
+      if ((comp = (regex_t *) malloc(sizeof *comp)) == NULL) {
+	 log_log(LOG_ERR, NULL,
+		 _("No more memory to store the denied IP addresses"));
+	 
+	 return -1;
+      }
+      if ((err = regcomp(comp, item, REG_EXTENDED | REG_ICASE)) != 0) {
+	 char errbuf[LINE_MAX + 1];
+	 
+	 regerror(err, comp, errbuf, sizeof LINE_MAX);
+	 log_log(LOG_ERR, NULL,
+		 _("Suspicious regex : [%s] <%s>"), item, errbuf);
+	 
+	 return -1;
+      }
+      currentservicecontext->rx_deny_ip = comp;
+   } else {
+      currentservicecontext->rx_deny_ip = NULL;
+   }
+
+   return 0;
+}
+
+PRO(DENY_IP_EXCEPT)
+{
+   char *item;
+   ServiceContext *currentservicecontext =
+     parsercontext->currentservicecontext;
+   regex_t *comp;
+   int err;
+   
+   if (GETNEXTOKEN) {
+      if ((comp = (regex_t *) malloc(sizeof *comp)) == NULL) {
+	 log_log(LOG_ERR, NULL,
+		 _("No more memory to store the excepted denied IP addresses"));
+
+	 return -1;
+      }
+      if ((err = regcomp(comp, item, REG_EXTENDED | REG_ICASE)) != 0) {
+	 char errbuf[LINE_MAX + 1];
+	 
+	 regerror(err, comp, errbuf, sizeof LINE_MAX);
+	 log_log(LOG_ERR, NULL,
+		 _("Suspicious regex : [%s] <%s>"), item, errbuf);
+	 
+	 return -1;
+      }
+      currentservicecontext->rx_deny_ip_except = comp;
+   } else {
+      currentservicecontext->rx_deny_ip_except = NULL;
+   }
+
+   return 0;
+}
+
 PRO(SERVICE_NBLIMIT)
 {
    char *item;
@@ -637,7 +906,31 @@ PRO(SERVICE_NBLIMIT)
 	 return -1;
       }
       parsercontext->currentservicecontext->service_nblimit =
-	(size_t) atol(item);
+	(size_t) parser_atol(item);
+   } else {
+      goto cr0;
+   }   
+   return 0;
+}
+
+PRO(SPAMFILTER)
+{
+   char *item;
+   int t;
+   
+   if (GETNEXTOKEN) {      
+      if (parser_checknumber(item) == 0) {	 
+	 cr0:
+	 log_log(LOG_ERR, NULL, 
+		 _("ERROR : SPAMFILTER needs a number : 0 to disable, 1 or 2 to enable"));
+	      
+	 return -1;
+      }
+      t = atoi(item);
+      if (t < 0 || t > 2) {
+	 goto cr0;
+      }
+      parsercontext->currentservicecontext->spamfilter = t;
    } else {
       goto cr0;
    }   
