@@ -16,6 +16,7 @@
 #endif                 
 
 extern int errno;
+extern int h_errno;
 
 /* Just a stupid malloc() wrapper */
 
@@ -107,16 +108,16 @@ static struct hostent *socket_gethostent(const char *hostname_,
 
 /* Find a port number */
 
-static unsigned short socket_getport(const char * const protocol,
-				     const char * const portname,
-				     const unsigned short port)
+static int socket_getport(const char * const protocol,
+			  const char * const portname,
+			  const int port)
 {
    struct servent *service;
 
    if ((service = getservbyname(portname, protocol)) != NULL) {
-      return service->s_port;
+      return (int) ntohs(service->s_port);
    }
-   return htons(port);
+   return port;
 }
 
 /* Find a protocol (default = def) */
@@ -134,18 +135,43 @@ static int socket_getproto(const char * const protoname, const int def)
 
 /* These options are systematically added to any socket */
 
-static int socket_setopts(const int fd)
+static int socket_setopts(const int fd, const int reuseaddr)
 {
-   socklen_t opt = 1;
-   
+   int opt = 1;
+
    setsockopt(fd, socket_getproto("ip", IPPROTO_IP), IP_OPTIONS, NULL, 0);
-   setsockopt(fd, SOL_SOCKET, SO_REUSEADDR, (char *) &opt, sizeof opt);
+   if (reuseaddr == 0) {
+      opt = 0;
+   } else {
+      opt = 1;
+   }
+   setsockopt(fd, SOL_SOCKET, SO_REUSEADDR, (char *) &opt, 
+	      (socklen_t) sizeof opt);
 #ifdef SO_USELOOPBACK
-   setsockopt(fd, SOL_SOCKET, SO_USELOOPBACK, (char *) &opt, sizeof opt);
+   opt = 1;
+   setsockopt(fd, SOL_SOCKET, SO_USELOOPBACK, (char *) &opt, 
+	      (socklen_t) sizeof opt);
 #endif
-   opt = 0;
 #ifdef SO_BROADCAST
-   setsockopt(fd, SOL_SOCKET, SO_BROADCAST, (char *) &opt, sizeof opt);
+   opt = 0;   
+   setsockopt(fd, SOL_SOCKET, SO_BROADCAST, (char *) &opt, 
+	      (socklen_t) sizeof opt);
+#endif
+#ifndef NO_LINGER
+# ifdef SO_LINGER
+   {
+      struct linger lng;
+#  ifdef BYPASS_LINGER
+      lng.l_onoff = 0;
+      lng.l_linger = 0;
+#  else      
+      lng.l_onoff = 1;
+      lng.l_linger = SOCKET_LINGER;
+#  endif
+      setsockopt(fd, SOL_SOCKET, SO_LINGER, (char *) &lng, 
+		 (socklen_t) sizeof lng);
+   }
+# endif
 #endif
    
    return 0;
@@ -168,12 +194,12 @@ static int socket_fixopts(Socket * const socket, const int fd)
       
    static unsigned char optbuf[MAX_IP_OPTIONS];
    Sockets * const sockets = socket->sockets;
-   socklen_t optbuflen = sizeof optbuf;
+   socklen_t optbuflen = (socklen_t) sizeof optbuf;
    size_t optlen = 0;
    unsigned char *optbufpnt = optbuf;
 
    if (getsockopt(fd, socket_getproto("ip", IPPROTO_IP), IP_OPTIONS,
-		  (char *) optbuf, &optbuflen) == 0 &&
+		  (char *) optbuf, (void *) &optbuflen) == 0 &&
        optbuflen != 0) {
       
 # ifdef FIRST_HOP_IN_GETSOCKOPT
@@ -282,6 +308,7 @@ static int socket_fixopts(Socket * const socket, const int fd)
 
 static int socket_create(const int endpoint,
 			 const struct protoent * const proto,
+			 const int reuseaddr,
 #ifdef IPV6_SUPPORT
 			 struct sockaddr_in6 * const local
 #else
@@ -298,21 +325,55 @@ static int socket_create(const int endpoint,
 #endif
 #ifdef IPV6_SUPPORT
    if ((fd = socket(PF_INET6, endpoint, proto->p_proto)) < 0 ||
-       socket_setopts(fd) < 0 ||
+       socket_setopts(fd, reuseaddr) < 0 ||
        bind(fd, (struct sockaddr *) local, sizeof *local) < 0) {
+      close(fd);
       return -1;
    }
 #else
    if ((fd = socket(PF_INET, endpoint, proto->p_proto)) < 0 ||   
-          socket_setopts(fd) < 0 ||
+          socket_setopts(fd, reuseaddr) < 0 ||
               bind(fd, (struct sockaddr *) local, sizeof *local) < 0) {
-            return -1; 
+      close(fd);
+      return -1; 
    }
 #endif
+   if (tools_set_cloexec_flag(fd, 1)) {
+   	  /* it's not OK for j.r. User to see the sockets of other users ! */
+	  /* Should we log this? */
+   	  close(fd);
+	  return -1;
+   }
       
    return fd;
 }
 
+/* Close a socket and discards its data */
+
+int socket_realclose(const int fd)
+{
+#ifdef SO_LINGER
+     {
+	struct linger lng;	
+# ifdef BYPASS_LINGER
+	lng.l_onoff = 0;
+	lng.l_linger = 0;
+# else      
+	lng.l_onoff = 1;
+	lng.l_linger = 0;
+# endif	
+	setsockopt(fd, SOL_SOCKET, SO_LINGER, (char *) &lng, 
+		   (socklen_t) sizeof lng);
+     }
+#endif
+#ifdef HAVE_SHUTDOWN
+   shutdown(fd, 2);
+#endif   
+   close(fd);
+   
+   return 0;
+}
+
 /* Return 1 if fd points to an IPv6 socket */
 
 int socket_isipv6(const int fd)
@@ -423,6 +484,19 @@ static int socket_listadd(Sockets * const sockets, Socket * const socket)
    return 0;
 }
 
+/* Reset the socket info */
+
+static void socket_resetsocketinfo(SockEvent * const sockevent)
+{
+   SocketInfo * const socketinfo = &sockevent->socketinfo;
+   
+   *socketinfo->protocol = 0;
+   *socketinfo->localport = *socketinfo->remoteport = 0;
+   *socketinfo->localhost = *socketinfo->remotehost = 0;
+   *socketinfo->remoteinfo = 0;
+   *socketinfo->localip = *socketinfo->remoteip = 0;
+}
+
 /* Create a new client socket */
 
 Socket * socket_clientcreate(Sockets * const sockets,			      
@@ -430,7 +504,7 @@ Socket * socket_clientcreate(Sockets * const sockets,
 			     const char * const hostnameenv,
 			     const char * const endpoint,			     
 			     const char * const protocol,
-			     const char * const portname, unsigned short port)
+			     const char * const portname, int port)
 {
    Socket *socket;
    struct hostent *h;
@@ -447,12 +521,21 @@ Socket * socket_clientcreate(Sockets * const sockets,
       return NULL;
    }
    if ((h = socket_gethostent(hostname, hostnameenv)) == NULL) {
+      errno = ENOENT;
       return NULL;
    }
    if ((socket->endpoint = parsendpoint(endpoint)) == 0) {
+#ifdef EPROTO
+      errno = EPROTO;
+#endif
+      return NULL;
+   }
+   if ((port = socket_getport(protocol, portname, port)) < 0) {
+# ifdef EPROTO
+      errno = EPROTO;
+# endif
       return NULL;
    }
-   port = socket_getport(protocol, portname, port);
    socket->port = port;
    memset(&server, 0, sizeof server);
    memset(&local, 0, sizeof local);
@@ -460,13 +543,13 @@ Socket * socket_clientcreate(Sockets * const sockets,
    server.sin6_family = local.sin6_family = AF_INET6;
    server.sin6_flowinfo = local.sin6_flowinfo = 0;
    local.sin6_addr = in6addr_any;
-   server.sin6_port = port;
-   local.sin6_port = htons(0);   
+   server.sin6_port = htons((unsigned short) port);
+   local.sin6_port = htons((unsigned short) 0);   
 #else
    server.sin_family = local.sin_family = AF_INET;
    local.sin_addr.s_addr = INADDR_ANY;
-   server.sin_port = port;
-   local.sin_port = htons(0);   
+   server.sin_port = htons((unsigned short) port);
+   local.sin_port = htons((unsigned short) 0);   
 #endif
    proto = getprotobyname(protocol);
    addrpnt = h->h_addr_list;
@@ -476,8 +559,8 @@ Socket * socket_clientcreate(Sockets * const sockets,
 #else
       memcpy(&server.sin_addr, *addrpnt, (size_t) h->h_length);
 #endif
-      if ((fd = socket_create(socket->endpoint, proto, &local)) < 0) {
-
+      if ((fd = socket_create(socket->endpoint, proto, 
+			      (port < SOCKET_MINUSERPORT), &local)) < 0) {
 	 return NULL;
       }
       socket->addr = local;
@@ -496,8 +579,8 @@ Socket * socket_clientcreate(Sockets * const sockets,
    socket->geek = socket->geek2 = NULL;
    socket->fd = fd;
    socket->optionnodelay = socket->optionkeepalive = SOCKET_OPTION_DEFAULT;
-
    sockets->fdtosocket[fd] = socket;
+   socket_resetsocketinfo(&sockets->sockevent);   
    if (fd >= sockets->lastfd) {
       if (fd < SOCKMAX_FD) {
 	 sockets->lastfd = fd + 1;
@@ -534,6 +617,7 @@ int socket_initsockets(Sockets * const sockets)
    }
    FD_ZERO(&sockets->lastrfds);      
    FD_ZERO(&sockets->rfdsbackup);
+   socket_resetsocketinfo(&sockets->sockevent);   
    sockets->sockevent.type = SOCKEVENT_NONE;
    sockets->pendingmessages = 0;
    sockets->pendingsync = SOCKET_PENDINGSYNC;
@@ -603,7 +687,7 @@ Socket * socket_servercreate(Sockets * const sockets,
 			     const char * const hostnameenv,
 			     const char * const endpoint,			     
 			     const char * const protocol,
-			     const char * const portname, unsigned short port)
+			     const char * const portname, int port)
 {
    Socket *socket;
    struct hostent *h;
@@ -623,20 +707,29 @@ Socket * socket_servercreate(Sockets * const sockets,
    socket->geek = socket->geek2 = NULL;
    socket->optionnodelay = socket->optionkeepalive = SOCKET_OPTION_DEFAULT;
    if ((socket->endpoint = parsendpoint(endpoint)) < 0) {
+#ifdef EPROTO
+      errno = EPROTO;
+#endif
+      return NULL;
+   }
+   if ((port = socket_getport(protocol, portname, port)) < 0) {
+#ifdef EPROTO
+      errno = EPROTO;
+#endif
       return NULL;
    }
-   port = socket_getport(protocol, portname, port);
    memset(&server, 0, sizeof server);
    memset(&local, 0, sizeof local);
 #ifdef IPV6_SUPPORT
-   local.sin6_port = port;   
+   local.sin6_port = htons((unsigned short) port);
 #else
-   local.sin_port = port;
+   local.sin_port = htons((unsigned short) port);
 #endif
    socket->port = port;
    proto = getprotobyname(protocol);
    if (hostname != NULL) {
       if ((h = socket_gethostent(hostname, hostnameenv)) == NULL) {
+	 errno = ENOENT;
 	 return NULL;
       }   
 #ifdef IPV6_SUPPORT
@@ -654,13 +747,15 @@ Socket * socket_servercreate(Sockets * const sockets,
    }
    socket->addr = local;
    if (socket->endpoint == SOCK_STREAM) {
-      if ((fd = socket_create(SOCK_STREAM, proto, &local)) < 0 || 
+      if ((fd = socket_create(SOCK_STREAM, proto, 
+			      (port < SOCKET_MINUSERPORT), &local)) < 0 || 
 	  listen(fd, SOCKET_BACKLOG) < 0) {
 	 
 	 return NULL;
       }   
    } else {
-      if ((fd = socket_create(socket->endpoint, proto, &local)) < 0) {
+      if ((fd = socket_create(socket->endpoint, proto,
+			      (port < SOCKET_MINUSERPORT), &local)) < 0) {
 	 
 	 return NULL;
       }
@@ -668,6 +763,7 @@ Socket * socket_servercreate(Sockets * const sockets,
    socket->fd = fd;   
    socket_listadd(sockets, socket);
    sockets->fdtosocket[fd] = socket;
+   socket_resetsocketinfo(&sockets->sockevent);      
    if (fd >= sockets->lastfd) {
       if (fd < SOCKMAX_FD) {
 	 sockets->lastfd = fd + 1;
@@ -728,6 +824,22 @@ int socket_hangup(Socket * const socket, const int fd)
    if (socket->endpoint != SOCK_STREAM) {
       return 0;
    }
+#ifndef SOFT_WITH_CLIENT
+# ifdef SO_LINGER
+     {
+	struct linger lng;	
+#  ifdef BYPASS_LINGER
+	lng.l_onoff = 0;
+	lng.l_linger = 0;
+#  else      
+	lng.l_onoff = 1;
+	lng.l_linger = 0;
+#  endif	
+	setsockopt(fd, SOL_SOCKET, SO_LINGER, (char *) &lng, 
+		   (socklen_t) sizeof lng);
+     }
+# endif
+#endif
    close(fd);
    FD_CLR(fd, &socket->sockets->rfdsbackup);
    if (fd == socket->sockets->lastfd - 1) {
@@ -770,7 +882,9 @@ char *socket_read(const int fd)
 int socket_flush(const int fd)
 {
    char tmpbuf[SOCKLINE_MAX];
-   recv(fd, tmpbuf, sizeof tmpbuf, 0);
+   int foo;
+   
+   recvfrom(fd, tmpbuf, sizeof tmpbuf, 0, NULL, &foo);
    
    return 0;
 }
@@ -817,8 +931,8 @@ int socket_servercnxwait(Socket * const socket)
    if (socket->endpoint != SOCK_STREAM) {
      return fd;
    }
-   if ((t = accept(fd, (struct sockaddr *) &remote, &addrlen)) < 0) {
-
+   if ((t = accept(fd, (struct sockaddr *) &remote, 
+		   (void *) &addrlen)) < 0) {
       return -1;
    }
    
@@ -827,30 +941,36 @@ int socket_servercnxwait(Socket * const socket)
     * fake packets. Drop the connexion. */
    
 #ifdef IPV6_SUPPORT
-   if (remote.sin6_port < SOCKET_MINUSERPORT ||
+   if (
+# ifndef SOCKET_BROKENPORT
+       ntohs(remote.sin6_port) < (unsigned short) SOCKET_MINUSERPORT ||
+# endif
        socket_addrcompare(&remote, &socket->addr) == 0) {
       close(t);
       
       return -2;
    }
 #else
-   if (remote.sin_port < SOCKET_MINUSERPORT ||
+   if (
+# ifndef SOCKET_BROKENPORT
+       ntohs(remote.sin_port) < (unsigned short) SOCKET_MINUSERPORT ||
+# endif
        socket_addrcompare(&remote, &socket->addr) == 0) {
       close(t);
-      
       return -2;
    }
 #endif
 
    socket_fixopts(socket, t);
-   if (socket->optionnodelay != SOCKET_OPTION_DEFAULT) {
+   if (socket->optionnodelay != SOCKET_OPTION_DEFAULT) {      
       int opt = socket->optionnodelay;
       setsockopt(t, socket_getproto("tcp", IPPROTO_TCP), TCP_NODELAY,
-		 (char *) &opt, sizeof opt);
+		 (char *) &opt, (socklen_t) sizeof opt);
    }
    if (socket->optionkeepalive != SOCKET_OPTION_DEFAULT) {
       int opt = socket->optionkeepalive;
-      setsockopt(t, SOL_SOCKET, SO_KEEPALIVE, (char *) &opt, sizeof opt);
+      setsockopt(t, SOL_SOCKET, SO_KEEPALIVE, (char *) &opt, 
+		 (socklen_t) sizeof opt);
    }
    
    return t;
@@ -892,6 +1012,7 @@ SockEvent *socket_serverwaitevent(Sockets * const sockets)
    } else if (sockets->pendingmessages < 0) {
       sockets->pendingmessages = 0;
    }
+   socket_resetsocketinfo(&sockets->sockevent);      
    rfds = sockets->rfdsbackup;
    if ((selectret = select(sockets->lastfd, &rfds, NULL, NULL, NULL)) <= 0) {
       if (errno == EINTR) {
@@ -945,7 +1066,44 @@ SockEvent *socket_serverwaitevent(Sockets * const sockets)
       sockets->sockevent.remotecnx[newcnxfd] = 1;
       sockets->fdtosocket[newcnxfd] = socket;
       sockets->sockevent.type = SOCKEVENT_NEWCNX;
-      
+#ifndef SOCKET_DONTGETINFO
+	{
+	   SocketInfo * const socketinfo = &sockets->sockevent.socketinfo;
+	   const SocketLocalInfo * const localinfo =
+	     socket_getlocalinfo(socket, newcnxfd);
+	   const char *tmp;
+	   const char **tmp2;
+	   int tmp3;
+
+	   if (localinfo != NULL) {
+	      if (localinfo->localip != NULL) {
+		 strncpy(socketinfo->localip, localinfo->localip, sizeof socketinfo->localip);
+	      }
+	      if (localinfo->localhost != NULL) {
+		 strncpy(socketinfo->localhost, localinfo->localhost, sizeof socketinfo->localhost);
+	      }
+	   }
+# ifdef HAVE_SNPRINTF
+	   snprintf(socketinfo->localport, sizeof socketinfo->localport, "%hd", socket->port);
+# else
+	   sprintf(socketinfo->localport, "%hd", socket->port);
+# endif
+	   if ((tmp3 = socket_getpeerport(socket, newcnxfd)) >= 0) {
+# ifdef HAVE_SNPRINTF
+	      snprintf(socketinfo->remoteport, sizeof socketinfo->localport, "%d", tmp3);
+# else
+	      sprintf(socketinfo->remoteport, "%d", tmp3);
+# endif	   
+	   }
+	   if ((tmp = socket_getpeeraddr(socket, newcnxfd)) != NULL) {
+	      strncpy(socketinfo->remoteip, tmp, sizeof socketinfo->remoteip);
+	   }
+	   if ((tmp2 = socket_getpeername(socket, newcnxfd)) != NULL && *tmp2 != NULL) {
+	      strncpy(socketinfo->remotehost, *tmp2, sizeof socketinfo->remotehost);
+	   }	   
+	}
+#endif
+	      
       return &sockets->sockevent;
    }
    scannedfd = 0;
@@ -988,10 +1146,10 @@ SockEvent *socket_serverwaitevent(Sockets * const sockets)
 /* Look up a remote host sockaddr info */
 
 #ifdef IPV6_SUPPORT
-static struct sockaddr_in6 *socket_getpeer(Socket * const socket,
+static struct sockaddr_in6 *socket_getpeer(const Socket * const socket,
 					   const int fd)
 #else
-static struct sockaddr_in *socket_getpeer(Socket * const socket,
+static struct sockaddr_in *socket_getpeer(const Socket * const socket,
 					  const int fd)
 #endif
 {
@@ -1003,18 +1161,87 @@ static struct sockaddr_in *socket_getpeer(Socket * const socket,
    socklen_t len = sizeof remote;
    
    if (!socket_isstream(socket) ||
-       getpeername(fd, (struct sockaddr *) &remote, &len) < 0) {
+       getpeername(fd, (struct sockaddr *) &remote, (void *) &len) < 0) {
       char tmpbuf[SOCKLINE_MAX + 1];
       
       len = sizeof remote;
       if (recvfrom(fd, tmpbuf, SOCKLINE_MAX, MSG_PEEK,
-		   (struct sockaddr *) &remote, &len) < 0) {	   
+		   (struct sockaddr *) &remote, (void *) &len) < 0) {
+#ifdef IPV6_SUPPORT
+	  remote.sin6_addr = NULL;
+#else
+	  remote.sin_addr.s_addr = 0U;
+#endif	  
+      }
+   }
+   if (
+#ifdef IPV6_SUPPORT
+       remote.sin6_addr == NULL
+#else
+       remote.sin_addr.s_addr == 0U
+#endif
+       ) {
+      if (getpeername(fd, (struct sockaddr *) &remote, (void *) &len) < 0) {
 	 return NULL;
       }
    }
+   
    return &remote;
 }
 
+/* Look up a local host sockaddr info */
+
+#ifdef IPV6_SUPPORT
+static struct sockaddr_in6 *socket_getlocal(Socket * const socket,
+					    const int fd)
+#else
+static struct sockaddr_in *socket_getlocal(Socket * const socket,
+					   const int fd)
+#endif
+{
+#ifdef IPV6_SUPPORT
+   static struct sockaddr_in6 local;
+#else
+   static struct sockaddr_in local;
+#endif
+   socklen_t len = sizeof local;
+   
+   if (!socket_isstream(socket) ||
+       getsockname(fd, (struct sockaddr *) &local, (void *) &len) < 0) {
+      char tmpbuf[SOCKLINE_MAX + 1];
+      
+      len = sizeof local;
+      if (recvfrom(fd, tmpbuf, SOCKLINE_MAX, MSG_PEEK,
+		   (struct sockaddr *) &local, (void *) &len) < 0) {	   
+	 return NULL;
+      }
+   }
+   return &local;
+}
+
+/* Look up a remote port (in hardware order) */
+
+int socket_getpeerport(const Socket * const socket, const int fd)
+{
+#ifdef IPV6_SUPPORT
+   struct sockaddr_in6 *remote;
+#else
+   struct sockaddr_in *remote;
+#endif
+   
+#ifdef IPV6_SUPPORT
+   if ((remote = socket_getpeer(fd)) == NULL) {
+      return -1;
+   }
+   return (int) ntohs(remote->sin6_port);
+#else
+   if ((remote = socket_getpeer(socket, fd)) == NULL) {
+      return -1;      
+   }
+   return (int) ntohs(remote->sin_port);
+#endif
+}
+
 /* Look up a remote host address in dots notation */
 
 const char *socket_getpeeraddr(Socket * const socket, const int fd)
@@ -1033,13 +1260,123 @@ const char *socket_getpeeraddr(Socket * const socket, const int fd)
    return inet_ntoa(remote->sin6_addr); /* FIXME */
 #else
    if ((remote = socket_getpeer(socket, fd)) == NULL ||       
-       remote->sin_addr.s_addr == 0) {
+       remote->sin_addr.s_addr == 0U) {
       return NULL;      
    }
    return inet_ntoa(remote->sin_addr);
 #endif
 }
 
+/* See if a remote host is banned by RBL maps - No IPv6 support yet AFAIK */
+
+int socket_ispammer(Socket * const socket, const int fd, const char * const rbl)
+{
+   static char tmp[MAXHOSTNAMELEN];
+   struct sockaddr_in *remote;
+   const unsigned char *uglyhook;
+   struct hostent *ent;
+
+   if (socket_isipv6(fd) != 0) {
+      return -2;
+   }
+   if ((remote = socket_getpeer(socket, fd)) == NULL ||       
+       remote->sin_addr.s_addr == 0) {
+      return -1;
+   }
+   uglyhook = (const unsigned char *) &remote->sin_addr.s_addr;
+#ifdef HAVE_SNPRINTF
+   snprintf(tmp, sizeof tmp, "%d.%d.%d.%d.%s.", 
+	    uglyhook[3], uglyhook[2], uglyhook[1], uglyhook[0],
+	    rbl);
+#else
+   sprintf(tmp, "%d.%d.%d.%d.%s.", 
+	   uglyhook[3], uglyhook[2], uglyhook[1], uglyhook[0],
+	   rbl);     
+#endif
+   ent = gethostbyname(tmp);
+   if (ent != NULL) {
+      return 1;
+   } else if (h_errno == HOST_NOT_FOUND
+#ifdef NO_ADDRESS
+	      || h_errno == NO_ADDRESS
+#endif
+#ifdef NO_DATA
+	      || h_errno == NO_DATA
+#endif	      
+	      ) {
+      return 0;
+   }
+   return -1;
+}
+
+/* Look up a local host address in dots notation */
+
+const SocketLocalInfo *socket_getlocalinfo(Socket * const socket, const int fd)
+{
+   static SocketLocalInfo localinfo;
+#ifdef IPV6_SUPPORT
+   struct sockaddr_in6 *local;
+#else
+   struct sockaddr_in *local;
+#endif
+   struct hostent *host;
+   
+#ifdef IPV6_SUPPORT
+   if ((local = socket_getlocal(fd)) == NULL ||       
+       local->sin6_addr == NULL) {
+      return NULL;
+   }
+   localinfo.localip = inet_ntoa(local->sin6_addr); /* FIXME */
+   host = gethostbyaddr((char *) &local->sin6_addr,
+			sizeof local->sin6_addr, AF_INET);
+   
+#else
+   if ((local = socket_getlocal(socket, fd)) == NULL ||       
+       local->sin_addr.s_addr == 0) {
+      return NULL;
+   }
+   localinfo.localip = inet_ntoa(local->sin_addr);
+   host = gethostbyaddr((char *) &local->sin_addr,
+			sizeof local->sin_addr, AF_INET);   
+#endif
+   if (host != NULL && host->h_name != NULL && *host->h_name != 0) {
+      localinfo.localhost = host->h_name;
+   } else {
+      localinfo.localhost = NULL;
+   }
+   
+   return &localinfo;
+}
+
+/* A long host name can be composed by digits, dots, [a-z], and
+ * "-" . Nothing else or it is not RFC-compliant.
+ * If TOLERATE_UNDERSCORES_IN_FQDN is defined, then "_" is a valid
+ * character, too. */
+
+static int socket_validhostname(const char * const name)
+{
+   register const char *namepnt;
+   int c;
+   
+   if ((namepnt = name) == NULL || *namepnt == 0) {
+      return -1;
+   }
+   while ((c = *namepnt) != 0) {
+      if (!isalpha(c) && !isdigit(c) && c != '.' && c != '-') {
+#ifdef TOLERATE_UNDERSCORES_IN_FQDN
+	 if (c != '_') {
+	    return 0;
+	 }
+#else
+	 return 0;	 
+#endif
+      }
+      namepnt++;
+   }
+   
+   return 1;
+}
+
 /* Do the so-called "paranoid" DNS check and return 0 if
  * the host is not lying */
 
@@ -1053,7 +1390,11 @@ static int socket_paranoidcheck(const struct sockaddr_in *host,
 {
    const struct hostent *host2;
    char * const *addrpnt;
-   
+   int maxaddrpnt = SOCKET_MAXHOSTALIASES;   
+
+   if (socket_validhostname(name) <= 0) {   
+      return -1;
+   }
 #ifdef IPV6_SUPPORT
    if ((host2 = gethostbyname2(name)) == NULL) {
       return -1;
@@ -1071,7 +1412,7 @@ static int socket_paranoidcheck(const struct sockaddr_in *host,
    if (strcasecmp(name, host2->h_name) != 0 &&
        strcasecmp(name, "localhost") != 0) {
       return -1;
-   }
+   }   
 #endif
    addrpnt = host2->h_addr_list;
    while (*addrpnt != NULL) {
@@ -1079,6 +1420,10 @@ static int socket_paranoidcheck(const struct sockaddr_in *host,
 		 sizeof host->sin_addr) == 0) {
 	 return 0;
       }
+      maxaddrpnt--;
+      if (maxaddrpnt <= 0) {
+	 return -1;
+      }
       addrpnt++;
    }
    
@@ -1096,7 +1441,7 @@ const char **socket_getpeername(Socket * const socket, const int fd)
    static const char *hostnames[SOCKET_MAXHOSTALIASES + 2];
    
    const char **hostnamespnt = hostnames;
-   size_t hostnamescnt = SOCKET_MAXHOSTALIASES;
+   int hostnamescnt = SOCKET_MAXHOSTALIASES;
 #ifdef IPV6_SUPPORT
    struct sockaddr_in6 *remote;
 #else
@@ -1129,7 +1474,7 @@ const char **socket_getpeername(Socket * const socket, const int fd)
    if (socket_paranoidcheck(remote, host->h_name) == 0) {
       *hostnamespnt++ = host->h_name;
       hostnamescnt--;
-      if (hostnamescnt == 0) {
+      if (hostnamescnt <= 0) {
 	 goto coolworld;
       }
    }
@@ -1138,7 +1483,7 @@ const char **socket_getpeername(Socket * const socket, const int fd)
       if (socket_paranoidcheck(remote, *aliases) == 0) {
 	 *hostnamespnt++ = *aliases;
 	 hostnamescnt--;
-	 if (hostnamescnt == 0) {
+	 if (hostnamescnt <= 0) {
 	    goto coolworld;
 	 }
       }
@@ -1156,7 +1501,7 @@ const char *socket_getpeernameoraddr(Socket * const socket, const int fd)
 {
    const char **results;
    
-   if ((results = socket_getpeername(socket, fd)) != NULL) {
+   if ((results = socket_getpeername(socket, fd)) != NULL && *results != NULL) {
       return *results;
    }
    return socket_getpeeraddr(socket, fd);
